@@ -12,7 +12,7 @@ use strict;
 
 use vars qw{
 	$KNOWN_BAD_DIALUP_RANGES $IP_IN_RESERVED_RANGE
-	$EXISTING_DOMAIN $IS_DNS_AVAILABLE
+	$EXISTING_DOMAIN $IS_DNS_AVAILABLE $VERSION
 };
 
 $EXISTING_DOMAIN = 'microsoft.com.';
@@ -21,6 +21,8 @@ $IP_IN_RESERVED_RANGE = undef;
 
 $IS_DNS_AVAILABLE = undef;
 
+$VERSION = 'bogus';     # avoid CPAN.pm picking up razor ver
+
 ###########################################################################
 
 BEGIN {
@@ -63,11 +65,11 @@ sub do_rbl_lookup {
 	dbg ("record found for $dom = $addr");
 
 	if ($addr ne '127.0.0.2' && $addr ne '127.0.0.3') {
-	  $self->test_log ("RBL check: found relay ".$dom.", type: ".$addr);
+	  $self->test_log ("RBL check: found ".$dom.", type: ".$addr);
 	} else {
 	  # 127.0.0.2 is the traditional boolean indicator, don't log it
 	  # 127.0.0.3 now also means "is a dialup IP"
-	  $self->test_log ("RBL check: found relay ".$dom);
+	  $self->test_log ("RBL check: found ".$dom);
 	}
 
 	$self->{$set}->{rbl_IN_As_found} .= $addr.' ';
@@ -168,18 +170,22 @@ sub razor_lookup {
     open (STDOUT, ">&STDERR");
   }
 
+  my $oldslash = $/;
+
   eval {
     require Razor::Client;
     require Razor::Agent;
     local ($^W) = 0;		# argh, warnings in Razor
-
-    my $rc = Razor::Client->new ($config, %options);
-    die "undefined Razor::Client\n" if (!$rc);
+    local ($/);			# argh, bugs in Razor
 
     local $SIG{ALRM} = sub { die "alarm\n" };
     alarm 10;
 
-    if ($Razor::Client::VERSION >= 1.12) {
+    my $rc = Razor::Client->new ($config, %options);
+    die "undefined Razor::Client\n" if (!$rc);
+
+    my $ver = $Razor::Client::VERSION;
+    if ($ver >= 1.12) {
       my $respary = $rc->check ('spam' => \@msg);
       # response can be "0" or "1". there can be many responses.
       # so if we get 5 responses, and one of them's 1, we
@@ -195,12 +201,14 @@ sub razor_lookup {
 
   if ($@) {
     if ($@ =~ /alarm/) {
-      warn "razor check timed out after $timeout secs.\n";
+      dbg ("razor check timed out after $timeout secs.");
     } else {
       warn ("razor check skipped: $! $@");
     }
   }
 
+  $/ = $oldslash;		# argh! pollution!
+
   # razor also debugs to stdout. argh. fix it to stderr...
   if ($Mail::SpamAssassin::DEBUG) {
     open (STDOUT, ">&OLDOUT");
@@ -224,6 +232,7 @@ sub load_resolver {
     $self->{res} = Net::DNS::Resolver->new;
     if (defined $self->{res}) {
       $self->{no_resolver} = 0;
+      $self->{res}->retry(1); # If it fails, it fails
     }
     1;
   };   #  or warn "eval failed: $@ $!\n";
@@ -253,6 +262,40 @@ sub lookup_mx {
   return $ret;
 }
 
+sub lookup_ptr {
+  my ($self, $dom) = @_;
+
+  return undef unless $self->load_resolver();
+  if ($self->{main}->{local_tests_only}) {
+    dbg ("local tests only, not looking up PTR");
+    return undef;
+  }
+
+  dbg ("looking up PTR record for '$dom'");
+  my $name = '';
+
+  eval {
+        my $query = $self->{res}->search($dom);
+        if ($query) {
+	  foreach my $rr ($query->answer) {
+	    if ($rr->type eq "PTR") {
+	      $name = $rr->ptrdname; last;
+	    }
+	  }
+        }
+
+  };
+  if ($@) {
+    # 71 == EX_OSERR.  PTR lookups are not supposed to crash and burn!
+    sa_die (71, "PTR lookup died: $@ $!\n");
+  }
+
+  dbg ("PTR for '$dom': '$name'");
+
+  # note: undef is never returned, unless DNS is unavailable.
+  return $name;
+}
+
 sub is_dns_available {
   my ($self) = @_;
 
