@@ -6,10 +6,11 @@ use Carp;
 use strict;
 
 use vars	qw{
-  	@ISA
+  	@ISA $VERSION
 };
 
 @ISA = qw();
+$VERSION = 'bogus';	# avoid CPAN.pm picking up razor ver
 
 ###########################################################################
 
@@ -50,10 +51,16 @@ sub report {
 
 sub is_razor_available {
   my ($self) = @_;
+
+  if ($self->{main}->{local_tests_only}) {
+    dbg ("local tests only, ignoring Razor");
+    return 0;
+  }
   
   eval {
     require Razor::Client;
   };
+
   if ($@) {
     dbg ( "Razor is not available" );
     return 0;
@@ -67,11 +74,12 @@ sub razor_report {
   my ($self, $fulltext) = @_;
 
   my @msg = split (/^/m, $fulltext);
+  my $timeout = 10;             # seconds
+  my $response;
   my $config = $self->{main}->{conf}->{razor_config};
   my %options = (
     'debug'     => $Mail::SpamAssassin::DEBUG
   );
-  my $response;
 
   # razor also debugs to stdout. argh. fix it to stderr...
   if ($Mail::SpamAssassin::DEBUG) {
@@ -79,28 +87,43 @@ sub razor_report {
     open (STDOUT, ">&STDERR");
   }
 
+  my $oldslash = $/;
+
   eval {
     require Razor::Client;
+    require Razor::Agent;
     local ($^W) = 0;            # argh, warnings in Razor
+    local ($/);                 # argh, bugs in Razor
+
+    local $SIG{ALRM} = sub { die "alarm\n" };
+    alarm 10;
 
     my $rc = Razor::Client->new ($config, %options);
     die "undefined Razor::Client\n" if (!$rc);
 
-    if ($Razor::Client::VERSION >= 1.12) {
+    my $ver = $Razor::Client::VERSION;
+    if ($ver >= 1.12) {
       my $respary = $rc->report ('spam' => \@msg);
       for my $resp (@$respary) { $response .= $resp; }
     } else {
       $response = $rc->report (\@msg);
     }
 
+    alarm 0;
     dbg ("Razor: spam reported, response is \"$response\".");
   };
   
   if ($@) {
-    warn "razor-report failed: $! $@";
+    if ($@ =~ /alarm/) {
+      dbg ("razor report timed out after $timeout secs.");
+    } else {
+      warn "razor-report failed: $! $@";
+    }
     undef $response;
   }
 
+  $/ = $oldslash;
+
   if ($Mail::SpamAssassin::DEBUG) {
     open (STDOUT, ">&OLDOUT");
     close OLDOUT;
